package se.umu.jobsubmission.client; 

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.RandomAccessFile;
import java.io.StringWriter;
import java.net.InetAddress;
import java.util.Calendar;
import java.util.List;
import java.util.Vector;

import javax.xml.rpc.Stub;

import org.apache.axis.AxisFault;
import org.apache.axis.message.addressing.Address;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.ggf.schemas.jsdl._2005._06.jsdl.JobDefinition_Type;
import org.globus.axis.util.Util;
import org.globus.gsi.GlobusCredential;
import org.globus.wsrf.client.ServiceURL;
import org.globus.wsrf.config.ContainerConfig;
import org.globus.wsrf.encoding.DeserializationException;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.encoding.ObjectSerializer;
import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.utils.FaultHelper;
import org.oasis.wsrf.faults.BaseFaultType;
import org.xml.sax.InputSource;

import se.umu.jobsubmission.Benchmark;
import se.umu.jobsubmission.ClientConfiguration;
import se.umu.jobsubmission.InvalidJobDescriptionFaultType;
import se.umu.jobsubmission.JobObjectiveType;
import se.umu.jobsubmission.JobPreferencesType;
import se.umu.jobsubmission.JobRequest;
import se.umu.jobsubmission.JobRequestFaultType;
import se.umu.jobsubmission.JobRequirementType;
import se.umu.jobsubmission.JobSubmission;
import se.umu.jobsubmission.JobSubmissionServiceAddressingLocator;
import se.umu.jobsubmission.NoAptResourcesFoundFaultType;
import se.umu.jobsubmission.NoResourcesFoundFaultType;
import se.umu.jobsubmission.RequestJob;
import se.umu.jobsubmission.RequestJobResponse;
import se.umu.jobsubmission.SchedulingHintsType;
import se.umu.jobsubmission.impl.datamanager.DataHelper;
import se.umu.jobsubmission.impl.jobsubmissionservice.JobSubmissionConstants;
import se.umu.jobsubmission.util.JobDescriptionTranslator;


/**
 * Simple client for job submission. 
 * The client accepts four parameters, 
 * the job description (mandatory),
 * a job preferences document (optional), 
 * a list of index servers (optional), 
 * and an URL to the JSS to use (optional)
 */

public class SubmitJob {

    static {
        Util.registerTransport(); 
    }

    private String jobDescriptionFile; 
    private String jobPreferencesFile; 
    private String indexServers[]; 
    private String URI; 
    private String host; 
    private String clientHost; 
    private boolean clientStagesInput; 
    
    private String[] converterClasses; 
    private String[] stagingTargetResolverClasses; 
    
    private JobDescriptionTranslator jobDescriptionTranslators[]; 

    private static final String optionsDesc = 
        "OPTIONS:\n" +
        "\n"+
        "Help:\n" +
        " -help                 display help.\n" +
        " -usage                display usage.\n" +
        "Usage:\n" + 
        " -f                    Specify file that contains job description\n" + 
        " -i                    Specify IndexService to retrieve resources from\n" + 
        " -j                    Specify URL of job submission service to contact\n" +
        " -p                    Specify file that contains job preferences\n" +
		" -s                    Disable client file staging\n" + 
        "\n"; 

    private static final String usageDesc =
        "ARGUMENTS: \n" +
        "         [options] -f <job description file name>\n" +
        "with\n" +
        "         [options]       [-i <indexService]+ \n" +
        "                         [-p <job preferences file name>]\n" +
        "                         [-j <job submission service> (default is localhost)]\n" +
		"                         [-s]\n" + 
        "\n";

    public SubmitJob() throws Exception {
    	
        this.jobDescriptionFile = null; 
        this.jobPreferencesFile = null; 
        // Using resolved address instead of 127.0.0.1 which gives 
        // some problems with authentication as certificates differs. 
        InetAddress address = InetAddress.getLocalHost();     
        this.host = address.getCanonicalHostName(); 
        this.clientHost = host; 
        
        this.URI = ServiceURL.getURL(
        		host, JobSubmissionConstants.SERVICE_NAME).
					toString(); 
        
        this.indexServers = null; 
        
        this.clientStagesInput = true; 
        
        // Read and deserialize config file 
    	String configFile = "/etc/jobsubmission/clientconfig.xml"; 
    	String globusLocation = ContainerConfig.getGlobusLocation(); 
    	String filePath = globusLocation + configFile; 
        
        ClientConfiguration config = null;
        
        FileInputStream input = new FileInputStream(filePath); 
        InputSource inputSource = new InputSource(input); 
        config = (ClientConfiguration) ObjectDeserializer.deserialize(inputSource, 
        		ClientConfiguration.class); 
        
        this.converterClasses = config.getDescriptionConverter(); 
        this.stagingTargetResolverClasses = config.getStagingTargetResolver(); 
        
    }
      
    /**
     * Parse user input
     * @param args
     * @throws Exception
     */
    public void processArguments(String args[]) throws Exception {

        List indexServersList = new Vector(); 

        if(args.length == 0) {
            System.out.println();
            System.out.println(usageDesc);
            System.out.println(optionsDesc);
            System.exit(0); 
        }

        boolean error = false; 
        for(int i = 0; i < args.length && !error; i++) {

            if (args[i].startsWith("-")) {

                if (args[i].equalsIgnoreCase("-help")) {
                    System.out.println();
                    System.out.println(usageDesc);
                    System.out.println(optionsDesc);
                    System.exit(0);
                }
                else if (args[i].equalsIgnoreCase("-usage")) {
                    System.out.println(usageDesc);
                    System.exit(0);
                }
                else if (args[i].equals("-f")) {
                    i++; 
                    if(i == args.length) {
                        System.err.println("-f requires a job description file");
                        error = true; 
                    }
                    else
                        jobDescriptionFile = args[i]; 
                }
                else if (args[i].equals("-i")) {
                    i++; 
                    if(i == args.length) {
                        System.err.println("-i requires a index service URI");
                        error = true; 
                    }
                    else 
                        indexServersList.add(args[i]); 
                }
                else if (args[i].equals("-j")) {
                    i++; 
                    if(i == args.length) {
                        System.err.println("-j requires a job submission service URI");
                        error = true; 
                    }
                    else  {
                        host = args[i]; 
                
                        URI = ServiceURL.getURL(
                        		host, 
								JobSubmissionConstants.SERVICE_NAME).toString();
                    }
							
                }
                else if (args[i].equals("-p")) {
                    i++; 
                    if(i == args.length) {
                        System.err.println("-p requires a job preferences file");
                        error = true; 
                    }
                    else
                        jobPreferencesFile = args[i]; 
                }
                else if (args[i].equals("-s")) {
                	System.out.println("File staging will be performed by "
                			+ "service, not by client"); 
                	this.clientStagesInput = false; 
                }
                else {
                    System.err.println("Unknown argument: " + args[i]);
                    error = true; 
                }
            }
            else {
                System.err.println("Unknown argument: " + args[i]);
                error = true; 
            }
        }
        if(error) {
            System.exit(1); 
        }

        int noIndexServers = indexServersList.size(); 

        if(noIndexServers > 0) {
            indexServers = new String[noIndexServers]; 
            for(int i = 0; i < noIndexServers; i++) 
                indexServers[i] = (String) indexServersList.get(i); 
        }
    }
    
    /**
     * Returns the list of job description translators as specified
     * in the client configuration file
     * @param clientConfig
     * @return
     */
    public JobDescriptionTranslator[] getConverter(String clientConfig) {
    	
    	try {
    		BufferedReader in = new BufferedReader(new FileReader(clientConfig));
    		List classNames = new Vector(); 
    		while(in.ready()) {
    			String className = in.readLine().trim();
    			classNames.add(className); 
    		}
    		in.close();
    		JobDescriptionTranslator converters[] = 
    			new JobDescriptionTranslator[classNames.size()]; 
    		
    		for(int i = 0; i < classNames.size(); i++) {
    			converters[i] = (JobDescriptionTranslator) 
				Class.forName((String)classNames.get(i)).newInstance(); 
    		}
    	
    		return converters; 
    	} catch (Exception e) {
    		System.err.println("Can not create converter:"); 
    		return null; 
    	}
    }

    /**
     * Convert job description to JSDL, 
     * the client config file determines which converter to use. 
     */
    public JobDefinition_Type convertJobDescription(String fileName) 
    throws FileNotFoundException, Exception {
    	  
    	// Try first for file existence
    	try {
    		RandomAccessFile f = new RandomAccessFile(fileName, "r");
    		f.close(); 
    	} catch (FileNotFoundException e) {
    		throw new FileNotFoundException ("Job description file " + fileName 
    				+ " not found"); 
    	}
    	
    	for(int i = 0; i < converterClasses.length; i++) {
    		JobDescriptionTranslator converter = null; 
    		try {
    			converter = (JobDescriptionTranslator) Class.forName(
    					this.converterClasses[i]).newInstance(); 
    			
    			JobDefinition_Type jobDef = (JobDefinition_Type) converter.
						translate(fileName); 
    			System.out.println("Translated job description " + fileName + 
    					" using " + converter.getClass().getName()); 
    			return jobDef; 
    		} catch (Exception e) {
    			System.out.println("Could not use " + converter.getClass().getName()
    					+ " to convert job description " + fileName); 
    		}
    	}
    	throw new Exception("No valid converter exists for " + fileName); 
    }
    
    /**
     * Reads and parses the job preferences file. If everything works
     * as expected, add the preferences file to the jobrequest. 
     * @param request
     * @param fileName
     */
    public void setJobPreferences(JobRequest request, String fileName) {
    	
    	if(this.jobPreferencesFile != null) {
        	
        	try {
        		JobPreferencesType preferences = null; 
        		FileInputStream input = new FileInputStream(fileName); 
        		InputSource inputSource = new InputSource(input); 
        		preferences = (JobPreferencesType) ObjectDeserializer.deserialize( 
        				inputSource, JobPreferencesType.class);  
        		
        		printPreferences(preferences); 
        		
        		request.setJobPreferences(preferences); 
        	} catch (FileNotFoundException e) {
        		System.err.println("Job preferences file " + 
        				this.jobPreferencesFile + 
						" not found, no job preferences will be used"); 
        	} catch(DeserializationException e) {
        		System.err.println("Deserialization error:\n" + e.getMessage()); 
        	} catch (Exception e) {
        		System.err.println("Could not read job preferences from file "
        				+ this.jobPreferencesFile); 
        	}
        }
    }
    
    /**
     * Performs client-side file staging. When doing this, 
     * the JSS will not stage any input files. Client side file staging is done 
     * unless the user explicitly requests otherwise. 
     * @param jobID
     * @param jobDesc
     * @throws Exception
     */
    public boolean stageJobInput(String jobID, JobDefinition_Type jobDesc) 
    throws Exception {
    	
    	List localInputFiles = null; 
    	
    	// Check if any job input files must be staged (at all) 
    	localInputFiles = DataHelper.getLocalInputFiles(jobDesc); 
    	if(localInputFiles == null) {
    		return false; 
    	}
    	
    	// Ok, must perform file staging. 
    	String filePath = null; 
    	
    	// Find out where to stage job input files. 
    	boolean done = false; 
    	for(int i = 0; i < this.stagingTargetResolverClasses.length && !done; i++) {
    		try {
    			ClientJobStaging stagingResolver = (ClientJobStaging) Class.forName(
    					this.stagingTargetResolverClasses[i]).newInstance(); 
    			filePath = stagingResolver.getJobDirectory(jobID, jobDesc); 
    			done = true; 
    		} catch (Exception e) {
    		}
    	}
    	
    	// Do not know where to stage input files -> cancel job. 
    	if(filePath == null) {
    		System.err.println("Can not determine where to stage input files"); 
    		throw new Exception("Can not determine where to stage input files"); 
    	}
    	
    	GlobusCredential cred = GlobusCredential.getDefaultCredential(); 
    	// Perform actual staging (cancel job if staging fails). 
    	DataHelper.stageLocalInputFiles(filePath, localInputFiles, 
    			this.clientHost, cred); 
    	
    	return true; 
    }
    
    /**
     * Cancel job if file staging failes. This is handled by the same
     * plugin that handles resolving of staging target URL. 
     * @param jobId
     */
    public void cancelJob(String jobId) {
    	boolean done = false; 
    	System.out.println("Cancelling job as file staging failed..."); 
    	for(int i = 0; i < this.stagingTargetResolverClasses.length && !done; i++) {
    		try {
    			ClientJobStaging stagingResolver = (ClientJobStaging) Class.forName(
    					this.stagingTargetResolverClasses[i]).newInstance(); 
    			stagingResolver.cancelJob(jobId); 
    			done = true; 
    			System.out.println("... job cancelled"); 
    		} catch (Exception e) {
    		}
    	}
    	if(!done) {
    		System.err.println("... can not cancel job"); 
    	}
    	
    }
     
    /**
     * Locate JobSubmissionService, convert job description
     * and submit it to service. Include any user preferences document
     * and/or non-default index server(s) in the job request. 
     * @throws Exception
     */
    public void submitJob() throws FileNotFoundException, Exception {

        System.out.println("Submitting job using:\n" + URI);

        System.out.println("Converting job description to JSDL");
        if(jobDescriptionFile == null) {
        	System.err.println("The -f flag (JobDescription) must be specified"); 
        	throw new 
				Exception("The -f flag (JobDescription) must be specified"); 
        }        
        
        JobDefinition_Type jobDesc = convertJobDescription(jobDescriptionFile);         

        System.out.println("Locating job submission service"); 
        // Locate service
        JobSubmissionServiceAddressingLocator locator = 
            new JobSubmissionServiceAddressingLocator(); 
                        
        JobSubmission jobsubmissionservice = null; 
        
        try { 
        	EndpointReferenceType endpoint = new EndpointReferenceType();
        	endpoint.setAddress(new Address(URI));
        	jobsubmissionservice = locator.getJobSubmissionPort(endpoint);        
        } catch (Exception e) {
        	System.err.println("Can not contact job submission service at " + URI); 
        	throw e; 
        }
        
        // Client security descriptor file 
        String clientDesc = "etc/jobsubmission/client-security-config.xml"; 
        // Set descriptor on Stub 
        ((Stub)jobsubmissionservice)._setProperty(
        		Constants.CLIENT_DESCRIPTOR_FILE, 
				clientDesc);
                
        // Build job request message
        RequestJob request = new RequestJob(); 
        JobRequest requestParameters = new JobRequest(); 

        // Index servers
        requestParameters.setIndexServerURIs(this.indexServers); 
        
        // JobDescription
        requestParameters.setJobDescription(jobDesc); 
                
        // Jobpreferences file 
        setJobPreferences(requestParameters, this.jobPreferencesFile); 
        
        // ClientHost       
        requestParameters.setClientURL(this.clientHost);
        
        // Indicate whether client stages job files. 
        requestParameters.setClientStagesFiles(new Boolean(this.clientStagesInput)); 
    
        request.setRequest(requestParameters);
        
        System.out.println("Submitting job..."); 
        RequestJobResponse response = null; 
      
        // Perform actual submission
        try {
        	response = jobsubmissionservice.requestJob(request); 
        }  
        catch (InvalidJobDescriptionFaultType f) {
        	System.err.println("Job request failed, invalid job description:"); 
        	FaultHelper helper = new FaultHelper(f); 
        	System.err.println(helper.getDescriptionAsString()); 
        } catch(NoResourcesFoundFaultType f) {
        	System.err.println("Job request failed"); 
        	FaultHelper helper = new FaultHelper(f); 
        	System.err.println(helper.getDescriptionAsString()); 
        } catch(NoAptResourcesFoundFaultType f) {
        	System.err.println("Job request failed"); 
        	FaultHelper helper = new FaultHelper(f); 
        	System.out.println(helper.getDescriptionAsString()); 
        } catch(JobRequestFaultType f) {
        	System.err.println("Job request failed:"); 
        	FaultHelper helper = new FaultHelper(f); 
        	System.err.println(helper.getDescriptionAsString()); 
        } catch (BaseFaultType f) {        	
        	System.err.println("General fault: "); 
        	FaultHelper helper = new FaultHelper(f);         	        	        	
        	System.err.println(helper.getDescriptionAsString());         	
        } catch(AxisFault f) {
        	System.err.println("Can not contact job submission service: " + URI); 
        	System.err.println("Reason: " + f.getFaultReason()); 
        } 
        
        if(response != null) {
        	try { 
        		if(this.clientStagesInput) {
        			System.out.println("Staging input files..."); 
        			if(stageJobInput(response.getJobId(), jobDesc)) 
        				System.out.println("... done"); 
        			else 
        				System.out.println("... no input files to stage"); 
        			System.out.println("Job submitted with jobid " 
        					+ response.getJobId()); 
        		}
        	} catch (Exception e) {
        		System.err.println("Error staging input files: " + e.getMessage());         		
        		cancelJob(response.getJobId()); 
        	}
        }
    }
    
    /**
     * Serializes job description and prints it to System.out.
     * This method is only used for debug purposes. 
     * @param job_desc
     */
    public void printJobDescription(JobDefinition_Type job_desc) throws Exception {
    	
    	StringWriter sw = new StringWriter(); 
    	ObjectSerializer.serialize(sw, 
    			job_desc, 
				JobDefinition_Type.getTypeDesc().getXmlType()); 
    	System.out.println("Job description:");
    	System.out.println(sw.toString()); 
    	
    }
    
    /**
     * Prints the job preferences, only used for debug purposes. 
     * @param prefs
     */
    public void printPreferences(JobPreferencesType prefs) {
    	    	
    	if(prefs == null) {
    		System.out.println("No preferences specified"); 
    		return; 
    	}
    	
    	SchedulingHintsType hints = prefs.getSchedulingHints(); 
    	
    	if(hints != null) {
    		System.out.println("Scheduling hints:");
    		Benchmark benchmarks[] = hints.getBenchmark(); 
    		if(benchmarks != null) {
    			int size = benchmarks.length; 
    			for(int i = 0; i < size; i++) {
    				System.out.println("Benchmark: " 
    						+ benchmarks[i].getName() + " "
							+ benchmarks[i].getResult() + " "
							+ benchmarks[i].getTime()); 
    			}
    		}
    	
    	    if(hints.getFileTransfers() != null) {
    		System.out.println("Filetransfer times : in / out: " +
    			hints.getFileTransfers().getStageInTime() +
				" / " + hints.getFileTransfers().getStageOutTime());
    	    }
    	}
    	
    	JobRequirementType reqs = prefs.getJobRequirement();
    	if(reqs != null) {
    		System.out.println("Job requirements:");    	    	
    		
    		System.out.print("EarliestStart:"); 
    		Calendar c = reqs.getEarliestAbsoluteStart(); 
    		if(c != null)    		
    			System.out.println(c.getTime());
    		else
    			System.out.println(reqs.getEarliestRelativeStart());    	
    		
    		System.out.print("LatestStart: "); 
			c = reqs.getLatestAbsoluteStart(); 
			if(c != null)
				System.out.println(c.getTime()); 
			else
				System.out.println(reqs.getLatestRelativeStart()); 
    		
			Boolean useRes = reqs.getUseReservations(); 
			if(useRes == null || useRes.booleanValue()) {
								
				Boolean flex = reqs.getFlexible();
				if(flex != null && flex.booleanValue())
					System.out.println("Flexible reservation requested");
				else 
					System.out.println("Fixed reservation requested"); 
			}
			else {
				System.out.println("No reservation requested"); 
			}
			
    		JobObjectiveType objective = reqs.getJobObjective(); 
    		if(objective.equals(JobObjectiveType.earliestJobStart)) {
    			System.out.println("Earliest possible job start requested"); 
    		}
    		else if(objective.equals(JobObjectiveType.earliestJobCompletion))
    			System.out.println("Earliest possible job completion requested"); 
    		
    		String targets[] = reqs.getSubmissionTarget(); 
    		if(targets != null) {
    			System.out.println("Possible submission targets:"); 
    			for(int i = 0; i < targets.length; i++)
    				System.out.println(targets[i]); 
    		}
    		
    	}
    	    	    	    
    }

    public static void main(String args[]) {

        try {

            SubmitJob s = new SubmitJob(); 
            s.processArguments(args); 
            s.submitJob(); 

        } catch (FileNotFoundException e) {
        	System.err.println(e.getMessage()); 
        }
        catch (Exception e) {
            System.out.println("Error submitting job");
        }
    }
}


