package se.umu.wsag.impl;  

import java.io.FileInputStream;
import java.net.URL;
import java.rmi.RemoteException;
import java.util.List;
import java.util.Vector;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.xml.namespace.QName;
import javax.xml.rpc.server.ServiceLifecycle;

import org.apache.axis.MessageContext;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.log4j.Logger;
import org.ggf.ws_agreement.AgreementPropertiesType;
import org.ggf.ws_agreement.AgreementTemplateType;
import org.ggf.ws_agreement.AgreementType;
import org.ggf.ws_agreement.ContinuingFaultType;
import org.ggf.ws_agreement.CreateAgreementInputType;
import org.ggf.ws_agreement.CreateAgreementOutputType;
import org.globus.wsrf.Resource;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourceProperties;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourcePropertySet;
import org.globus.wsrf.config.ContainerConfig;
import org.globus.wsrf.container.ServiceHost;
import org.globus.wsrf.encoding.ObjectDeserializer; 
import org.globus.wsrf.impl.ReflectionResourceProperty;
import org.globus.wsrf.impl.SimpleResourceKey;
import org.globus.wsrf.impl.SimpleResourcePropertySet;
import org.globus.wsrf.utils.AddressingUtils;
import org.globus.wsrf.utils.FilePersistenceHelper;
import org.oasis.wsrf.faults.BaseFaultTypeDescription;
import org.oasis.wsrf.properties.ResourceUnknownFaultType;
import org.xml.sax.InputSource; 

import se.umu.reservation.AgreementFactoryConfig;
import se.umu.reservation.DecisionMakerConfig;
import se.umu.wsag.util.JobSubmissionUtils;


/** AgreementFactory implemented using the singleton pattern. 
 * No resources are created, RPs are stored within the service. 
 */
public class AgreementFactoryService implements Resource, 
	ResourceProperties, ServiceLifecycle {
	
	private static Logger logger = 
		Logger.getLogger(AgreementFactoryService.class);
	
	private String agreementServiceURI;
	private AgreementFactoryConfig factoryConfig;
	private AgreementDecisionMaker decisionMaker;    
	private AgreementResourceHome agreementHome; 
	private AgreementResourceHelper resourceHelper; 
	
	// Resource properties
	private ResourcePropertySet rpSet = null;  
	// private AgreementFactoryPropertiesType template;
	private AgreementTemplateType template;    
	
	// FIXME: Specify paths as filename-only in config file, 
	// add globus-location + service-etc-dir when reading config. 
	// 
	/** Constructs configuration object from configFile.
	 * For each template file name in config, 
	 * parse agreement template and makes a RP of it. 
	 * Also determines which DecisionMaker to use, and initializes
	 * this DecisionMaker from configuration file 
	 */
	public void configure(String configFile) throws Exception {
		
		// First, create AgreementFactoryConfig object from 
		// configfile
		
		AgreementFactoryConfig config = null;  
		QName configQName = AgreementFactoryConfig.getTypeDesc().getXmlType();
		FileInputStream input = new FileInputStream(configFile);     	
		try { 
		InputSource inputSource = new InputSource(input); 
		config = (AgreementFactoryConfig) ObjectDeserializer.deserialize(
				inputSource, 
				AgreementFactoryConfig.class); 
		
		} catch (Exception e) {
			logger.error("Ooops: "); 
			e.printStackTrace(); 
		}
		
		// Set up which AgreementDecisionMaker to use: 
		DecisionMakerConfig decisionMakerConfig = config.
				getDecisionMakerConfig();    	    	
		String decisionClassName = decisionMakerConfig.getClassName();     	    	
		this.decisionMaker = (AgreementDecisionMaker) Class.forName
				(decisionClassName).newInstance();
		// Config params = decisionMakerConfig.getConfig(); 
		logger.debug("Initializing decisionMaker"); 
		this.decisionMaker.init(decisionMakerConfig); 
		logger.debug("DecisionMaker is: " +	decisionMaker.getClass());
		
		// Set up which AgreementResourceHelper to use:
		String agreementResourceHelperClass = config.
		getAgreementResourceHelperClass(); 
		this.resourceHelper = (AgreementResourceHelper) Class.forName
		(agreementResourceHelperClass).newInstance(); 
		logger.debug("ResourceHelper is: " + resourceHelper.getClass());  
		
		
		// Set up which agreement templates to use:      	
		// parse each agreement.
		String fileNames[] = config.getTemplateFileName();
		for(int i = 0; i < fileNames.length; i++)
			logger.debug("TemplateFile: " + fileNames[i]);
		
		String globusLocation = ContainerConfig.getGlobusLocation();
		List templatesList = new Vector();
		QName templateQName = AgreementTemplateType.getTypeDesc().getXmlType(); 
		for(int i = 0; i < fileNames.length; i++) {
			String filePath = globusLocation + fileNames[i];     		
			AgreementTemplateType template; 
			FileInputStream templateInput = new FileInputStream(filePath);
			InputSource templateInputSource = new InputSource(templateInput); 
			try {
				template = (AgreementTemplateType) ObjectDeserializer.
				deserialize(templateInputSource, AgreementTemplateType.class); 
				
				templatesList.add(template);
			} catch (Exception e) {
				logger.error("Error reading template: " + filePath);
				logger.error(e.getMessage()); 				
				throw e; 
			}
			
		}
		AgreementTemplateType templates[] = 
			new AgreementTemplateType[templatesList.size()];
		for(int i = 0; i < templatesList.size(); i++)
			templates[i] = (AgreementTemplateType) templatesList.get(i); 
		
		logger.debug("Inserting " + templatesList.size() + " templates into "
				+ "resource properties document."); 
		
		// Set up RP  
		// FIXME: Must be able to have arbitrarly many templates.
		// Stub containts only one template -> axis bug?  
		this.template = (AgreementTemplateType) templatesList.get(0);    	      
		this.rpSet = new SimpleResourcePropertySet(
				AgreementFactoryConstants.AGREEMENT_FACTORY_RP);
		
		ResourceProperty templateRP = new ReflectionResourceProperty(
				AgreementFactoryConstants.TEMPLATE_RP, "template", this);
		this.rpSet.add(templateRP);
		
	}
	
	/**
	 * AgreementFactoryService Instantiates RPs as this class is implemented
	 * using the Singleton pattern. 
	 * @throws RemoteException
	 */       
	public AgreementFactoryService() throws RemoteException {
		
		logger.debug("Creating AgreementFactoryService (singleton)"); 
		
		try {
			
			// Set up which instance class to create resources for:  
			URL baseURL = ServiceHost.getBaseURL();
			String instanceService = (String) MessageContext.
					getCurrentContext().getService().getOption("instance");
			agreementServiceURI = baseURL.toString() + instanceService;
			
			// Get hold of Agreement's resource home      
			this.agreementHome = (AgreementResourceHome)
			getJNDIObject("AgreementService", "/home"); 		            
			
			// Get location of template configuration file
			String configFile = (String) MessageContext. 
					getCurrentContext().getService().getOption("configFile");       
			String globusLocation = ContainerConfig.getGlobusLocation();             
			String filePath = globusLocation + configFile;			
			logger.debug("Configuring factory from file: " + filePath);
			
			// Read config file and 
			// i) create (deserialize) templates.
			// ii) create AgreementDecisionMaker
			configure(filePath);         	
			
		} catch (Exception e) {
			logger.error("Could not create AgreementFactoryService " 
					+ e.getMessage()); 			
			throw new  RuntimeException(
					"Could not create AgreementFactoryService", e); 
		}
	}
	
	// RP getters/setters
	public AgreementTemplateType getTemplate() {
		logger.debug("Returning template"); 
		return this.template; 
	}   
	public void setTemplate(AgreementTemplateType template) {
		logger.debug("Setting template"); 
		this.template = template; 
	}   
	
	/**
	 * The createAgreement operation is specified in the AgreementFactoryService
	 * porttype. Upon invocation, admission control is performed by the
	 * configured decisionmaker. If successful, an AgreementResource
	 * modelling the agreement is created, and the EPR of this 
	 * resource is returned. 
	 * @param params The input, contains the agreement offer. 
	 * @return
	 * @throws ResourceUnknownFaultType
	 * @throws RemoteException
	 * @throws ContinuingFaultType
	 */
	public CreateAgreementOutputType createAgreement(CreateAgreementInputType 
			params) throws ResourceUnknownFaultType, 
			RemoteException, 
			ContinuingFaultType {
		
		try {
			
			// Let AgreementDecision perform admission control
			AgreementType offer = params.getAgreementOffer();    		
			AgreementDecision decision = decisionMaker.create(offer);			
			
			// Admission control say 'ok' -> Create Agreement Resource
			// and store any domain-specific information 
			// (AgreemendDecision.context) in RPs of Agreement Resource.  
			if(decision.isDecision()) {    		
				
				// logger.debug("Creating agreement using 
				// AgreementFactoryService");
				
				// Create resource
				ResourceKey key = agreementHome.create(offer, 
						decision.getContext(), decisionMaker, resourceHelper);     		
				
				// logger.debug("AgreementResource created..."); 
				
				// logger.debug("Getting instance endpoint");                  
				URL baseURL = ServiceHost.getBaseURL();
				String instanceService = (String) MessageContext
				.getCurrentContext().getService().getOption("instance");
				String instanceURI = baseURL.toString() + instanceService;
				
				EndpointReferenceType epr =
					AddressingUtils.createEndpointReference(instanceURI, key);
				
				CreateAgreementOutputType res = new CreateAgreementOutputType();
				res.setCreatedAgreementEPR(epr); 
				
				return res;
			}
			else {
				ContinuingFaultType error = new ContinuingFaultType();
				BaseFaultTypeDescription descs[] = {new 
						BaseFaultTypeDescription("Agreement rejected due to: "
								+ decision.getReason())};     			    
				error.setDescription(descs); 
				throw error;     		
			}
			
		} catch (Exception e) {
			logger.debug("Creation failed..." + e.getMessage()); 
			throw new RemoteException("Creation failed " + e.getMessage(), e); 
		}
		
	}
	
	/**
	 * getJNDIObject Reads the jndi config file for this service 
	 * returns the specified object. Code extracted from MMJFS 
	 * @param service
	 * @param objectName
	 * @return
	 * @throws NamingException
	 */
	public static Object getJNDIObject(String serviceName, String objectName)
	throws NamingException {
		String jndiLookupString	= null;
		Object jndiObject;	
		try {
			Context initialContext = new InitialContext();
			if (!objectName.startsWith("/")) {
				objectName = "/" + objectName;
			}
			
			// Want string of format 
			// java:comp/env/services/ _SERVICENAME_ /home			
			// -> /home is object name...
			jndiLookupString = "java:comp/env/services/" + 
				serviceName + objectName;  
			
			// jndiLookupString = getServiceJNDILocation(service) + objectName;
			// jndiLookupString = serviceBase + objectName;
			// logger.debug("Searching for: " + jndiLookupString); 
			jndiObject = initialContext.lookup(jndiLookupString);
		} catch (NamingException e) {
			/*
			 String errorMessage = i18n.getMessage(Resources.JNDI_CONFIG_ERROR,
			 jndiLookupString);
			 logger.error(errorMessage + "\n" + e.getMessage());*/
			logger.error("Error looking up: " + jndiLookupString + " in jndi"); 
			throw e;
		}
		return jndiObject;
	}
	
	
	// Specified by ResourceProperties
	public ResourcePropertySet getResourcePropertySet() {
		return this.rpSet;
	}
	
	public void init(Object context) {	
		
		try {
			FilePersistenceHelper helper =  new FilePersistenceHelper(
					AgreementPropertiesType.class, ".conf");
			List resources = helper.list(); 
			logger.debug("Init, loading " + resources.size() + 
					" resource(s)..."); 
			Object keyObj = null; 
			for(int i = 0; i < resources.size(); i++) {
				try {
					logger.debug("ResourceKey: " + resources.get(i)); 
					keyObj = resources.get(i); 
					ResourceKey key = new SimpleResourceKey(
							agreementHome.getKeyTypeName(), keyObj); 
					// this should load the resource
					Resource r = this.agreementHome.find(key); 
				} catch (Exception e) {
					// The sweeper thread removed the resource
					// before we finished loading it...
				}
			}
		} catch (Exception e) {
			logger.debug("Error loading resource");
		}
		logger.debug("... done"); 
		
	}
	
	public void destroy() {
		logger.debug("Destroy...");
	}
	
}



/*            
 ResourceContext ctx = ResourceContext.getResourceContext();
 Iterator iter = ctx.getPropertyNames(); 
 while(iter.hasNext()) {
 String propertyName = (String) iter.next(); 
 Object property = ctx.getProperty(propertyName);             
 logger.debug("Property: " + propertyName + " " + property);
 }*/
